/* This file is part of the wvWare 2 project
   Copyright (C) 2001 Werner Trobin <trobin@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifndef OLESTORAGE_H
#define OLESTORAGE_H

extern "C"
{
#include <libole2/libole2.h>
}
#include <string>
#include <list>

namespace wvWare
{

class OLEStream;
class OLEStreamReader;
class OLEStreamWriter;

class OLEStorage
{
    friend class OLEStream;
public:
    /**
     * Create an "empty" storage
     */
    OLEStorage();
    /**
     * Specify a name for the storage. Note: It will *not* be opened
     * right now, call @see open() to do that
     */
    OLEStorage( const std::string& fileName );
    /**
     * Destroy the current storage. Open streams on it will
     * be synced and written back.
     */
    ~OLEStorage();

    /**
     * Specify a name for the storage (file)
     * This is a no-op if a storage is already open!
     */
    void setName( const std::string& fileName );
    /**
     * Get the name of the OLE storage file
     */
    std::string name() const { return m_fileName; }

    /**
     * Open the specified storage. If the specified file
     * doesn't exist we simply create one (it's safe to ignore the
     * warning on the console). Opening a storage twice won't do any harm
     * @return true if opening was successful
     */
    bool open();

    /**
     * Closes the current storage (no-op if we don't have one ;)
     * This method also tries to close/sync all the open streams.
     * Closing a storage twice won't do any harm
     */
    void close();

    /**
     * Any problems with the current storage?
     */
    bool isValid() const { return m_storage && m_lastError == MS_OLE_ERR_OK; }
    /**
     * The last error we received
     */
    MsOleErr lastError() const { return m_lastError; }

    /**
     * Returns a list containing all the files and dirs at the
     * current position in the VFS
     */
    std::list<std::string> listDirectory();

    /**
     * Enters a specific subdirectory. If the specified directory
     * doesn't exist this method returns false
     */
    bool enterDirectory( const std::string& directory );
    /**
     * One level up (cd ..)
     */
    void leaveDirectory();

    /**
     * Manipulate the path directly. If the operation fails at some
     * point the path will be set back to the previous location.
     */
    bool setPath( const std::string& path );
    /**
     * Get the current path
     */
    std::string path() const { return m_path; }

    /**
     * Stat an OLE node (in the current directory) to find out whether
     * it's a stream or a directory, find out about its size,...
     */
    MsOleStat stat( const std::string& node );

    /**
     * Opens the specified stream for reading and passes the
     * opened stream reader back. Returns 0 if it didn't work.
     * Note: The ownership is transferred to you!
     */
    OLEStreamReader* createStreamReader( const std::string& stream );

    /**
     * Opens a stream for writing (you get 0 if it failed)
     * Note: The ownership is transferred to you!
     * Note2: Don't try to pass names with a '/' in it :)
     */
    OLEStreamWriter* createStreamWriter( const std::string& stream );

    /**
     *Calls ms_ole_debug() on the current VFS (if there is any)
     */
    void debug() const;

private:
    /**
     * we don't want to allow copying
     */
    OLEStorage( const OLEStorage& rhs );
    /**
     * we don't want to allow assigning
     */
    OLEStorage& operator=( const OLEStorage& rhs );

    /**
     * This method is called from the stream's DTOR
     */
    void streamDestroyed( OLEStream* stream );

    MsOle* m_storage;    // our storage
    MsOleErr m_lastError;

    std::string m_fileName;
    std::string m_path;

    /**
     * We're not the owner, but we still keep track of all
     * the streams for bookkeeping issues
     */
    std::list<OLEStream*> m_streams;
};

} // namespace wvWare

#endif // OLESTORAGE_H
