/* This file is part of the wvWare 2 project
   Copyright (C) 2001 Werner Trobin <trobin@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "parser97.h"
#include "olestream.h"
#include "word97_helper.h"
#include "styles.h"
#include "textconverter.h"
#include "handlers.h"
#include "properties97.h"
#include "fonts.h"
#include "lists.h"
#include "footnotes97.h"
#include "headers97.h"
#include "fields.h"
#include "paragraphproperties.h"
#include "functor.h"
#include "functordata.h"
#include "associatedstrings.h"

#include <numeric>
#include <iostream>
#include <string.h>

using namespace wvWare;


Parser97::Position::Position( U32 cp, const Word97::PLCF<Word97::PCD>* plcfpcd ) :
    piece( 0 ), offset( cp )
{
    Word97::PLCFIterator<Word97::PCD> it( *plcfpcd );
    for ( ; it.current(); ++it, ++piece ) {
        if ( it.currentLim() > cp && it.currentStart() <= cp )
            break;
        offset -= it.currentRun();
    }
}


Parser97::Parser97( OLEStorage* storage, OLEStreamReader* wordDocument ) :
    Parser( storage, wordDocument ), Parser9x(), m_table( 0 ), m_data( 0 ), m_properties( 0 ),
    m_fonts( 0 ), m_lists( 0 ), m_footnotes( 0 ), m_headers( 0 ), m_fields( 0 ),
    m_magicTables( 0 ), m_fib( wordDocument, true ), m_plcfpcd( 0 ), m_textconverter( 0 ),
    m_tableRowStart( 0 ), m_tableRowLength( 0 ), m_cellMarkFound( false ), m_remainingCells( 0 ),
    m_currentParagraph( new Paragraph ), m_remainingChars( 0 ), m_sectionNumber( 0 ),
    m_subDocument( None ), m_parsingMode( Default )
{
    if ( !isOk() )
        return;

    // which table stream?
    m_table = storage->createStreamReader( m_fib.fWhichTblStm ? "1Table" : "0Table" );
    if ( !m_table || !m_table->isValid() ) {  // Are Word97 documents w/o table stream possible?
        std::cerr << "Error: Couldn't open the [0|1]Table stream" << std::endl;
        m_okay = false;                      // I don't think so
        return;
    }

    m_data = storage->createStreamReader( "Data" );
    if ( !m_data || !m_data->isValid() ) {
        std::cerr << "Warning: Couldn't open the Data stream, no big deal" << std::endl;
        delete m_data;
        m_data = 0;
    }

#ifdef WV2_DUMP_FIB
    std::cerr << "Dumping some parts of the FIB: " << std::endl;
    std::cerr << "   wIdent=" << m_fib.wIdent << std::endl;
    std::cerr << "   nFib=" << m_fib.nFib << std::endl;
    std::cerr << "   nFibBack=" << m_fib.nFibBack << std::endl;
    std::cerr << "   lid=0x" << std::hex << m_fib.lid << std::dec << std::endl;
    std::cerr << "   lidFE=0x" << std::hex << m_fib.lidFE << std::dec << std::endl;
    std::cerr << "   fEncrypted=" << m_fib.fEncrypted << std::endl;
    std::cerr << "   chs=" << m_fib.chs << std::endl;
    std::cerr << "   fcMin=" << m_fib.fcMin << std::endl;
    std::cerr << "   fcMac=" << m_fib.fcMac << std::endl;
    std::cerr << "   ccpText=" << m_fib.ccpText << std::endl;
    std::cerr << "   ccpFtn=" << m_fib.ccpFtn << std::endl;
    std::cerr << "   ccpHdd=" << m_fib.ccpHdd << std::endl;
    std::cerr << "   ccpMcr=" << m_fib.ccpMcr << std::endl;
    std::cerr << "   ccpAtn=" << m_fib.ccpAtn << std::endl;
    std::cerr << "   ccpEdn=" << m_fib.ccpEdn << std::endl;
    std::cerr << "   ccpTxbx=" << m_fib.ccpTxbx << std::endl;
    std::cerr << "   ccpHdrTxbx=" << m_fib.ccpHdrTxbx << std::endl;
    std::cerr << "   pnFbpChpFirst=" << m_fib.pnFbpChpFirst << std::endl;
    std::cerr << "   pnChpFirst=" << m_fib.pnChpFirst << std::endl;
    std::cerr << "   cpnBteChp=" << m_fib.cpnBteChp << std::endl;
    std::cerr << "   pnFbpPapFirst=" << m_fib.pnFbpPapFirst << std::endl;
    std::cerr << "   pnPapFirst=" << m_fib.pnPapFirst << std::endl;
    std::cerr << "   cpnBtePap=" << m_fib.cpnBtePap << std::endl;
#endif

    // Initialize all the cached data structures like stylesheets, fonts,
    // textconverter,...
    init();
}

Parser97::~Parser97()
{
    // Sanity check
    if ( !oldParsingStates.empty() || m_subDocument != None )
        std::cerr << "Bug: Someone messed up the save/restore stack!" << std::endl;

    delete m_currentParagraph;
    delete m_tableRowStart;
    delete m_textconverter;
    delete m_plcfpcd;
    delete m_magicTables;
    delete m_fields;
    delete m_headers;
    delete m_footnotes;
    delete m_lists;
    delete m_fonts;
    delete m_properties;
    delete m_data;
    delete m_table;
}

bool Parser97::parse()
{
    if ( !isOk() )
        return false;

    if ( m_fib.fEncrypted ) {
        // There is some code out there to break this "encryption", do we want
        // to implement that?
        // We could either ask for a password or cheat a bit :-)
        std::cerr << "Error: The document is encrypted." << std::endl;
        return false;
    }

    // Note that we test for the presence of complex structure, rather than
    // m_fib.fComplex. This allows us to treat newer files which always seem
    // to have piece tables in a consistent manner.
    //
    // There is also the implication that without the complex structures, the
    // text cannot be in unicode form.
    if ( m_fib.lcbClx != 0 ) {
        // Get the piece table
        if ( !readPieceTable() )
            return false;

        // start parsing the body
        if ( !parseBody() )
            return false;
    }
    else {
        std::cerr << "Bug: Found a Word97 file without piece table!" << std::endl;
        return false;
    }
    return true;
}

const Word97::FIB& Parser97::fib() const
{
    return m_fib;
}

const Word97::DOP& Parser97::dop() const
{
    return m_properties->dop();
}

const Word97::FFN& Parser97::font( S16 ftc ) const
{
    return m_fonts->font( ftc );
}

AssociatedStrings Parser97::associatedStrings()
{
    return AssociatedStrings( m_fib.fcSttbfAssoc, m_fib.lcbSttbfAssoc,
                              m_fib.fFarEast ? m_fib.lidFE : m_fib.lid, m_table );
}

const StyleSheet& Parser97::styleSheet() const
{
    return m_properties->styleSheet();
}

void Parser97::parseHeaders( const HeaderData& data )
{
    m_subDocumentHandler->headersStart();
    for ( unsigned char mask = 1; mask <= HeaderData::FooterFirst; mask <<= 1 )
        if ( mask & data.headerMask )
            parseHeader( data, mask );
    m_subDocumentHandler->headersEnd();
}

void Parser97::parseFootnote( const FootnoteData& data )
{
#ifdef WV2_DEBUG_FOOTNOTES
    std::cerr << "Parser97::parseFootnote() #####################" << std::endl;
#endif
    if ( data.limCP - data.startCP == 0 ) // shouldn't happen, but well...
        return;

    saveState( data.limCP - data.startCP, data.type == FootnoteData::Footnote ? Footnote : Endnote );
    m_subDocumentHandler->footnoteStart();

    U32 offset = m_fib.ccpText + data.startCP;
    if ( data.type == FootnoteData::Endnote )
        offset += m_fib.ccpFtn + m_fib.ccpHdd + m_fib.ccpMcr + m_fib.ccpAtn;
    parseHelper( Position( offset, m_plcfpcd ) );

    m_subDocumentHandler->footnoteEnd();
    restoreState();
#ifdef WV2_DEBUG_FOOTNOTES
    std::cerr << "Parser97::parseFootnote() done ################" << std::endl;
#endif
}

void Parser97::parseTableRow( const TableRowData& data )
{
#ifdef WV2_DEBUG_TABLES
    std::cerr << "Parser97::parseTableRow(): startPiece=" << data.startPiece << " startOffset="
              << data.startOffset << " length=" << data.length << std::endl;
#endif

    if ( data.length == 0 ) // idiot safe ;-)
        return;

    saveState( data.length, m_subDocument, Table );
    m_remainingCells = data.tap->itcMac;
    m_tableHandler->tableRowStart( data.tap );
    m_tableHandler->tableCellStart();

    parseHelper( Position( data.startPiece, data.startOffset ) );

    m_tableHandler->tableRowEnd();
    restoreState();

#ifdef WV2_DEBUG_TABLES
    std::cerr << "Parser97::parseTableRow() done #####################" << std::endl;
#endif
}

void Parser97::init()
{
    if ( m_fib.fFarEast )
        m_textconverter = new TextConverter( m_fib.lidFE );
    else
        m_textconverter = new TextConverter( m_fib.lid );

    // Get hold of all the SEP/PAP/CHP related structures and the StyleSheet
    m_properties = new Properties97( m_wordDocument, m_table, m_fib );
    m_fonts = new FontCollection( m_table, m_fib );
    m_lists = new ListInfoProvider( m_table, m_fib, &m_properties->styleSheet() );
    // Stage 2 of stylesheet construction, now that we know the list information
    // (chicken-egg problem)
    m_properties->styleSheet().createListInfo( *m_lists );

    if ( m_fib.ccpFtn != 0 )
        m_footnotes = new Footnotes97( m_fib, m_table );
    if ( m_fib.ccpHdd != 0 )
        m_headers = new Headers97( m_fib.fcPlcfhdd, m_fib.lcbPlcfhdd, m_table );

    m_fields = new Fields( m_fib, m_table );

    if ( m_fib.cfclcb > 93 && m_fib.lcbMagicTable != 0 ) {
        m_table->push();
        m_table->seek( m_fib.fcMagicTable );
        m_magicTables = new Word97::PLCF<Word97::BTE>( m_fib.lcbMagicTable, m_table );
        m_table->pop();
        /*
        std::cerr << "ALLLLLLLLOOOOOOOOOOOOAHHH" << std::endl;
        Word97::PLCFIterator<Word97::BTE> it( *m_magicTables );
        for ( ; it.current(); ++it )
            std::cerr << "start: " << it.currentStart() << " lim: " << it.currentLim()
                      << " entry: " << it.current()->pn << std::endl;
        std::cerr << "ALLLLLLLLOOOOOOOOOOOOAHHH" << std::endl;
        */
    }
}

bool Parser97::readPieceTable()
{
    m_table->seek( m_fib.fcClx );
    // first skip the leading grpprl blocks, we'll re-read them
    // if we need them later (no caching here)
    U8 blockType = m_table->readU8();
    while ( blockType == Word97::clxtGrpprl ) {
        U16 size = m_table->readU16();
#if WV2_DUMP_PIECE_TABLE > 0
        std::cerr << "Found a clxtGrpprl (size=" << size << ")" << std::endl;
#endif
        m_table->seek( size, SEEK_CUR );
        blockType = m_table->readU8();
    }
    if ( blockType == Word97::clxtPlcfpcd ) {
        U32 size = m_table->readU32();
#if WV2_DUMP_PIECE_TABLE > 0
        std::cerr << "Found the clxtPlcfpcd (size=" << size << ")" << std::endl;
#endif
        m_plcfpcd = new Word97::PLCF<Word97::PCD>( size, m_table, false );

#if WV2_DUMP_PIECE_TABLE > 1
        Word97::PLCFIterator<Word97::PCD> it( *m_plcfpcd );
        for ( int i = 0; it.current(); ++it, ++i ) {
            std::cerr << "Piece Table Entry(" << i << "): " << std::endl;
            std::cerr << "   start: " << it.currentStart() << std::endl;
            std::cerr << "   lim: " << it.currentLim() << std::endl;
            std::cerr << "   complex: " << it.current()->prm.fComplex << std::endl;
            if ( it.current()->prm.fComplex )
                std::cerr << "   igrpprl: " << it.current()->prm.toPRM2().igrpprl << std::endl;
            else
                std::cerr << "   isprm: " << it.current()->prm.isprm << std::endl;

            U32 fc = it.current()->fc;
            U32 limit = it.currentRun() << 1;
            std::cerr << "   value: " << fc << std::endl;
            if ( fc & 0x40000000 ) {
                fc = ( fc & 0xbfffffff ) >> 1;
                limit >>= 1;
                std::cerr << "   value (cleared 2nd MSB, div. by 2): " << fc << std::endl;
            }
            m_wordDocument->seek( fc );
            std::cerr << "   position: " << m_wordDocument->tell() << ", limit: " << limit << std::endl;
            for ( unsigned int j = 0; j < limit; ++j ) {
                U8 foo = m_wordDocument->readU8();
                if ( foo > 31 )
                    std::cerr << (char)foo;
                else if ( foo == PARAGRAPH_MARK )
                    std::cerr << std::endl;
                else if ( foo > 0 )
                    std::cerr << "{" <<  (int)foo << "}";
                else
                    std::cerr << "_";
            }
            std::cerr << std::endl << "   position: " << m_wordDocument->tell() << ", limit: " << limit << std::endl;
        }
#endif
    }
    else {
        std::cerr << "Oooops, couldn't find the piece table." << std::endl;
        return false;
    }
    return true;
}

bool Parser97::parseBody()
{
    saveState( m_fib.ccpText, Main );
    m_subDocumentHandler->bodyStart();

    SharedPtr<const Word97::SEP> sep( m_properties->sepForCP( 0 ) );
    if ( !sep )
        sep = new Word97::SEP(); // don't pass 0 pointers in any case
    m_textHandler->sectionStart( sep ); // First section, starting at CP 0
    emitHeaderData( sep );
    sep = 0; // get rid of the huge SEP

    // Process all the pieces belonging to the main document text
    parseHelper( Position( 0, static_cast<U32>( 0 ) ) );

    // Implicit end of the section
    m_textHandler->sectionEnd();
    m_subDocumentHandler->bodyEnd();
    restoreState();
    return true;
}

void Parser97::parseHelper( Position startPos )
{
    Word97::PLCFIterator<Word97::PCD> it( m_plcfpcd->at( startPos.piece ) );

    while ( m_remainingChars > 0 && it.current() ) {
        U32 fc = it.current()->fc;   // Start FC of this piece
        bool unicode;
        realFC( fc, unicode );

        U32 limit = it.currentRun(); // Number of characters in this piece

        // Check whether the text starts somewhere within the piece, reset at the end of the loop body
        if ( startPos.offset != 0 ) {
            fc += unicode ? startPos.offset * 2 : startPos.offset;
            limit -= startPos.offset;
        }

        limit = limit > m_remainingChars ? m_remainingChars : limit;
        m_wordDocument->seek( fc );

        if ( unicode ) {
            XCHAR* string = new XCHAR[ limit ];
            // First read the whole piece
            for ( unsigned int j = 0; j < limit; ++j ) {
                string[ j ] = m_wordDocument->readU16();
                if ( ( string[ j ] & 0xff00 ) == 0xf000 ) {
                    // Microsoft uses a Private Unicode Area (PUA) to store the characters of the
                    // Symbol and the Wingdings font. We simply clear these bits to shift the
                    // characters to 0x00XX and hope the correct font is installed. If the font
                    // isn't there, the user will get some ASCII text instead of symbols :}
                    //std::cerr << "private unicode area detected -- cropping" << std::endl;
                    string[ j ] &= 0x00ff;
                }
            }
            processPiece<XCHAR>( string, fc, limit, startPos ); // also takes care to delete [] string
        }
        else {
            U8* string = new U8[ limit ];
            m_wordDocument->read( string, limit );
            processPiece<U8>( string, fc, limit, startPos ); // also takes care to delete [] string
        }
        m_remainingChars -= limit;
        ++it;
        ++startPos.piece;
        startPos.offset = 0; // just in case it was != 0 in the first iteration
    }
}

template<typename String>
void Parser97::processPiece( String* string, U32 fc, U32 limit, const Position& position )
{
    // Take a closer look at the piece we just read. "start" and "index" are
    // counted in character positions (take care!)
    unsigned int start = 0;
    unsigned int index = 0;
    while ( index < limit ) {
        switch( string[ index ] ) {
            case SECTION_MARK:
            {
                if ( !m_currentParagraph->empty() || start != index ) {
                    // No "index - start + 1" here, as we don't want to copy the section mark!
                    UString ustring( processPieceStringHelper( string, start, index ) );
                    m_currentParagraph->push_back( Chunk( ustring, Position( position.piece, position.offset + start ),
                                                          fc + start * sizeof( String ), sizeof( String ) == sizeof( XCHAR ) ) );
                    processParagraph( fc + index * sizeof( String ) );
                }
                start = ++index;

                SharedPtr<const Word97::SEP> sep( m_properties->sepForCP( m_fib.ccpText - m_remainingChars + index ) );
                if ( sep ) {
                    // It's not only a page break, it's a new section
                    m_textHandler->sectionEnd();
                    m_textHandler->sectionStart( sep );
                    emitHeaderData( sep );
                }
                else
                    m_textHandler->pageBreak();
                break;
            }
            case CELL_MARK: // same ASCII code as a ROW_MARK
                m_cellMarkFound = true;
                // Fall-through intended. A row/cell end is also a paragraph end.
            case PARAGRAPH_MARK:
            {
                // No "index - start + 1" here, as we don't want to copy the paragraph mark!
                UString ustring( processPieceStringHelper( string, start, index ) );
                m_currentParagraph->push_back( Chunk( ustring, Position( position.piece, position.offset + start ),
                                                      fc + start * sizeof( String ), sizeof( String ) == sizeof( XCHAR ) ) );
                processParagraph( fc + index * sizeof( String ) );
                m_cellMarkFound = false;
                start = ++index;
                break;
            }
            // "Special" characters
            case TAB:
                string[ index ] = m_inlineHandler->tab();
                ++index;
                break;
            case HARD_LINE_BREAK:
                string[ index ] = m_inlineHandler->hardLineBreak();
                ++index;
                break;
            case COLUMN_BREAK:
                string[ index ] = m_inlineHandler->columnBreak();
                ++index;
                break;
            case NON_BREAKING_HYPHEN:
                string[ index ] = m_inlineHandler->nonBreakingHyphen();
                ++index;
                break;
            case NON_REQUIRED_HYPHEN:
                string[ index ] = m_inlineHandler->nonRequiredHyphen();
                ++index;
                break;
            case NON_BREAKING_SPACE:
                string[ index ] = m_inlineHandler->nonBreakingSpace();
                ++index;
                break;
            default:
                ++index;
                break;
        }
    }
    if ( start < limit ) {
        // Finally we have to add the remaining text to the current paragaph (if there is any)
        UString ustring( processPieceStringHelper( string, start, limit ) );
        m_currentParagraph->push_back( Chunk( ustring, Position( position.piece, position.offset + start ),
                                              fc + start * sizeof( String ), sizeof( String ) == sizeof( XCHAR ) ) );
    }
    delete [] string;
}

UString Parser97::processPieceStringHelper( XCHAR* string, unsigned int start, unsigned int index ) const
{
    return UString( reinterpret_cast<const wvWare::UChar *>( &string[ start ] ), index - start );
}

UString Parser97::processPieceStringHelper( U8* string, unsigned int start, unsigned int index ) const
{
    return m_textconverter->convert( reinterpret_cast<char*>( &string[ start ] ), index - start );
}

void Parser97::processParagraph( U32 fc )
{
    // Get the PAP structure as it was at the last full-save
    ParagraphProperties* props( m_properties->fullSavedPap( fc ) );
    // ...and apply the latest changes, then the PAP is completely restored
    m_properties->applyClxGrpprl( m_plcfpcd->at( m_currentParagraph->back().m_position.piece ).current(), m_fib.fcClx, props );

    // Skim the tables first, as soon as the functor is invoked we have to
    // parse them and emit the text
    if ( m_parsingMode == Default && props->pap().fInTable ) {
        if ( !m_tableRowStart ) {
            m_tableRowStart = new Position( m_currentParagraph->front().m_position );
            m_tableRowLength = 0;
#ifdef WV2_DEBUG_TABLES
            std::cerr << "Start of a table row: piece=" << m_tableRowStart->piece << " offset="
                      << m_tableRowStart->offset << std::endl;
#endif
        }
        m_tableRowLength += std::accumulate( m_currentParagraph->begin(), m_currentParagraph->end(),
                                             1, &Parser97::accumulativeLength ); // init == 1 because of the parag. mark!
        if ( props->pap().fTtp ) {
            // Restore the table properties of this row
            Word97::TAP* tap = m_properties->fullSavedTap( fc );
            m_properties->applyClxGrpprl( m_plcfpcd->at( m_currentParagraph->back().m_position.piece ).current(),
                                          m_fib.fcClx, tap, m_properties->styleByIndex( props->pap().istd ) );

            SharedPtr<const Word97::TAP> sharedTap( tap );
            // We decrement the length by 1 that the trailing row mark doesn't emit
            // one empty paragraph during parsing.
            m_tableHandler->tableRowFound( make_functor( *this, &Parser97::parseTableRow,
                                                         TableRowData( m_tableRowStart->piece, m_tableRowStart->offset,
                                                                       m_tableRowLength - 1, sharedTap ) ) );
            delete m_tableRowStart;
            m_tableRowStart = 0;
        }
    }
    else {
        // Now that we have the complete PAP, let's see if this paragraph belongs to a list
        props->createListInfo( *m_lists );

        SharedPtr<const ParagraphProperties> sharedProps( props ); // keep it that way, else the ParagraphProperties get deleted!
        m_textHandler->paragraphStart( sharedProps );

        // Get the appropriate style for this paragraph
        const Style* style = m_properties->styleByIndex( props->pap().istd );
        if ( !style ) {
            std::cerr << "Huh, really obscure error, couldn't find the Style for the current PAP -- skipping" << std::endl;
            return;
        }

        // Now walk the paragraph, chunk for chunk
        std::list<Chunk>::const_iterator it = m_currentParagraph->begin();
        std::list<Chunk>::const_iterator end = m_currentParagraph->end();
        for ( ; it != end; ++it ) {
            U32 index = 0;
            const U32 limit = ( *it ).m_text.length();
            const Word97::PLCFIterator<Word97::PCD> pcdIt( m_plcfpcd->at( ( *it ).m_position.piece ) );

            while ( index < limit ) {
                Word97::CHP* chp = new Word97::CHP( style->chp() );
                U32 length = m_properties->fullSavedChp( ( *it ).m_startFC + index * ( ( *it ).m_isUnicode ? 2 : 1 ), chp, style );
                if ( ( *it ).m_isUnicode )
                    length >>= 1;
                length = length > limit - index ? limit - index : length;

                m_properties->applyClxGrpprl( pcdIt.current(), m_fib.fcClx, chp, style );
                SharedPtr<const Word97::CHP> sharedChp( chp ); // keep it that way, else the CHP gets deleted!
                processChunk( *it, chp, length, index, pcdIt.currentStart() );
                index += length;
            }
        }
        m_textHandler->paragraphEnd();

        if ( m_cellMarkFound ) {
            m_tableHandler->tableCellEnd();
            if ( --m_remainingCells )
                m_tableHandler->tableCellStart();
        }
    }
    m_currentParagraph->clear();
}

void Parser97::processChunk( const Chunk& chunk, SharedPtr<const Word97::CHP> chp,
                             U32 length, U32 index, U32 currentStart )
{
    // Some characters have a special meaning (e.g. a footnote is anchored at some
    // position inside the text) and they *don't* have the fSpec flag set. This means
    // that we have to watch out for such characters even in plain text. Slooow :}
    //
    // For now we only have to handle footnote and endnote references that way. Due to that
    // the code below is a bit simpler right now, but I fear we have to extend that later on.
    // (We will have to keep track of the type of disruption, footnote() takes care of all now)
    //
    // A precondition for the footnote/endnote implementation below is, that footnote and
    // endnote references only occur in the main body text. The reason is that we only check
    // for the next footnote inside the PLCF and don't take subdocuments into account. If
    // it turns out that this precondition is not satisfied we would have to change the
    // O(1) nextFootnote() call to something like an O(n) containsFootnote( start, lim )
    // Up to now Word 97, 2000, and 2002 seem to be bug compatible and fullfill that precondition.
    //
    while ( length > 0 ) {
        U32 disruption = 0xffffffff; // "infinity"
        if ( m_footnotes ) {
            U32 nextFtn = m_footnotes->nextFootnote();
            U32 nextEnd = m_footnotes->nextEndnote();
            disruption = nextFtn < nextEnd ? nextFtn : nextEnd;
#ifdef WV2_DEBUG_FOOTNOTES
            std::cerr << "nextFtn=" << nextFtn << " nextEnd=" << nextEnd << " disruption="
                      << disruption << " length=" << length << std::endl;
#endif
        }
        U32 startCP = currentStart + chunk.m_position.offset + index;

        if ( disruption >= startCP && disruption < startCP + length ) {
#ifdef WV2_DEBUG_FOOTNOTES
            std::cerr << "startCP=" << startCP << " len=" << length << " disruption=" << disruption << std::endl;
#endif
            U32 disLen = disruption - startCP;
            if ( disLen != 0 )
                processRun( chunk, chp, disLen, index, currentStart );
            length -= disLen;
            index += disLen;
            processFootnote( chunk.m_text[ index ], disruption, chp );
            --length;
            ++index;
        }
        else {
            // common case, no disruption at all (or the end of a disrupted chunk)
            processRun( chunk, chp, length, index, currentStart );
            break;   // should be faster than messing with length...
        }
    }
}

void Parser97::processRun( const Chunk& chunk, SharedPtr<const Word97::CHP> chp,
                           U32 length, U32 index, U32 currentStart )
{
    if ( chp->fSpec ) {
        U32 i = 0;
        while ( i < length ) {
            processSpecialCharacter( chunk.m_text[ index + i ], currentStart + chunk.m_position.offset + index + i, chp );
            ++i;
        }
    }
    else {
        UConstString str( const_cast<UChar*>( chunk.m_text.data() ) + index, length );
        m_textHandler->runOfText( str.string(), chp );
    }
}

void Parser97::processSpecialCharacter( UChar character, U32 globalCP, SharedPtr<const Word97::CHP> chp )
{
    switch( character.unicode() ) {
        // Is it one of the "simple" special characters?
        case TextHandler::CurrentPageNumber:
            m_textHandler->specialCharacter( TextHandler::CurrentPageNumber, chp );
            break;
        case TextHandler::LineNumber:
            m_textHandler->specialCharacter( TextHandler::LineNumber, chp );
            break;
        case TextHandler::AbbreviatedDate:
            m_textHandler->specialCharacter( TextHandler::AbbreviatedDate, chp );
            break;
        case TextHandler::TimeHMS:
            m_textHandler->specialCharacter( TextHandler::TimeHMS, chp );
            break;
        case TextHandler::CurrentSectionNumber:
            m_textHandler->specialCharacter( TextHandler::CurrentSectionNumber, chp );
            break;
        case TextHandler::AbbreviatedDayOfWeek:
            m_textHandler->specialCharacter( TextHandler::AbbreviatedDayOfWeek, chp );
            break;
        case TextHandler::DayOfWeek:
            m_textHandler->specialCharacter( TextHandler::DayOfWeek, chp );
            break;
        case TextHandler::DayShort:
            m_textHandler->specialCharacter( TextHandler::DayShort, chp );
            break;
        case TextHandler::HourCurrentTime:
            m_textHandler->specialCharacter( TextHandler::HourCurrentTime, chp );
            break;
        case TextHandler::HourCurrentTimeTwoDigits:
            m_textHandler->specialCharacter( TextHandler::HourCurrentTimeTwoDigits, chp );
            break;
        case TextHandler::MinuteCurrentTime:
            m_textHandler->specialCharacter( TextHandler::MinuteCurrentTime, chp );
            break;
        case TextHandler::MinuteCurrentTimeTwoDigits:
            m_textHandler->specialCharacter( TextHandler::MinuteCurrentTimeTwoDigits, chp );
            break;
        case TextHandler::SecondsCurrentTime:
            m_textHandler->specialCharacter( TextHandler::SecondsCurrentTime, chp );
            break;
        case TextHandler::AMPMCurrentTime:
            m_textHandler->specialCharacter( TextHandler::AMPMCurrentTime, chp );
            break;
        case TextHandler::CurrentTimeHMSOld:
            m_textHandler->specialCharacter( TextHandler::CurrentTimeHMSOld, chp );
            break;
        case TextHandler::DateM:
            m_textHandler->specialCharacter( TextHandler::DateM, chp );
            break;
        case TextHandler::DateShort:
            m_textHandler->specialCharacter( TextHandler::DateShort, chp );
            break;
        case TextHandler::MonthShort:
            m_textHandler->specialCharacter( TextHandler::MonthShort, chp );
            break;
        case TextHandler::YearLong:
            m_textHandler->specialCharacter( TextHandler::YearLong, chp );
            break;
        case TextHandler::YearShort:
            m_textHandler->specialCharacter( TextHandler::YearShort, chp );
            break;
        case TextHandler::AbbreviatedMonth:
            m_textHandler->specialCharacter( TextHandler::AbbreviatedMonth, chp );
            break;
        case TextHandler::MonthLong:
            m_textHandler->specialCharacter( TextHandler::MonthLong, chp );
            break;
        case TextHandler::CurrentTimeHMS:
            m_textHandler->specialCharacter( TextHandler::CurrentTimeHMS, chp );
            break;
        case TextHandler::DateLong:
            m_textHandler->specialCharacter( TextHandler::DateLong, chp );
            break;

        // It has to be one of the very special characters...
        case TextHandler::FootnoteAuto:
            if ( m_subDocument == Footnote || m_subDocument == Endnote )
                m_textHandler->footnoteAutoNumber( chp );
            else
                processFootnote( character, globalCP, chp );
            break;
        case TextHandler::FieldBegin:
            m_textHandler->fieldStart( m_fields->fldForCP( m_subDocument, toLocalCP( globalCP ) ), chp );
            break;
        case TextHandler::FieldSeparator:
            m_textHandler->fieldSeparator( m_fields->fldForCP( m_subDocument, toLocalCP( globalCP ) ), chp );
            break;
        case TextHandler::FieldEnd:
            m_textHandler->fieldEnd( m_fields->fldForCP( m_subDocument, toLocalCP( globalCP ) ), chp );
            break;
        case TextHandler::FieldEscapeChar:
            std::cerr << "Found an escape character ++++++++++++++++++++?" << std::endl;
            break;
        default:
            std::cerr << "Parser97::processSpecialCharacter(): Support for character " << character.unicode()
                      << " not implemented yet." << std::endl;
            break;
    }
}

void Parser97::processFootnote( UChar character, U32 globalCP, SharedPtr<const Word97::CHP> chp )
{
    if ( !m_footnotes ) {
        std::cerr << "Bug: Found a footnote, but m_footnotes == 0!" << std::endl;
        return;
    }
#ifdef WV2_DEBUG_FOOTNOTES
    std::cerr << "######### Footnote found: CP=" << globalCP << std::endl;
#endif
    bool ok;
    FootnoteData data( m_footnotes->footnote( globalCP, ok ) );
    if ( ok )
        m_textHandler->footnoteFound( data.type, character, chp, make_functor( *this, &Parser97::parseFootnote, data ) );
}

void Parser97::emitHeaderData( SharedPtr<const Word97::SEP> sep )
{
    // We don't care about non-existant headers
    if ( !m_headers )
        return;

    // MS Word stores headers in a very strange way, so we have to keep track
    // of the section numbers. We use a 0-based index for convenience inside
    // the header reading code. (Werner)
    HeaderData data( m_sectionNumber++ );

    if ( sep->fTitlePage )
        data.headerMask |= HeaderData::HeaderFirst | HeaderData::FooterFirst;
    if ( dop().fFacingPages )
        data.headerMask |= HeaderData::HeaderEven | HeaderData::FooterEven;

    m_textHandler->headersFound( make_functor( *this, &Parser97::parseHeaders, data ) );
}

void Parser97::parseHeader( const HeaderData& data, unsigned char mask )
{
#ifdef WV2_DEBUG_HEADERS
    std::cerr << "parsing one header for section " << data.sectionNumber << ": mask=0x"
              <<  std::hex << ( int )mask << std::dec << std::endl;
#endif

    // First we have to determine the CP start/lim for the header text. From what I
    // found out Word 8 does it that way:
    //    - At the begin of the plcfhdd there are always 6 "0 fields" (stoppers)
    //    - The number of headers modulo 6 is always 0
    std::pair<U32, U32> range( m_headers->findHeader( data.sectionNumber, mask ) );

    int length = range.second - range.first;
#ifdef WV2_DEBUG_HEADERS
    std::cerr << "found a range: start=" << range.first << " lim=" << range.second << std::endl
              << "length: " << length << std::endl;
#endif
    if ( length < 1 ) {
#ifdef WV2_DEBUG_HEADERS
        std::cerr << "Warning: Didn't find a valid CPs for this header -- faking it" << std::endl;
#endif
        m_subDocumentHandler->headerStart( static_cast<HeaderData::Type>( mask ) );
        SharedPtr<const ParagraphProperties> sharedProps( new ParagraphProperties );
        m_textHandler->paragraphStart( sharedProps );
        m_textHandler->paragraphEnd();
        m_subDocumentHandler->headerEnd();
        return;
    }
    else if ( length > 1 )
        --length; // get rid of the trailing "end of header/footer" character

    saveState( length, Header );

    m_subDocumentHandler->headerStart( static_cast<HeaderData::Type>( mask ) );
    parseHelper( Position( m_fib.ccpText + m_fib.ccpFtn + range.first, m_plcfpcd ) );
    m_subDocumentHandler->headerEnd();

    restoreState();
}

void Parser97::saveState( U32 newRemainingChars, SubDocument newSubDocument, ParsingMode newParsingMode )
{
    oldParsingStates.push( ParsingState( m_tableRowStart, m_tableRowLength, m_remainingCells, m_currentParagraph,
                                         m_remainingChars, m_sectionNumber, m_subDocument, m_parsingMode ) );
    m_tableRowStart = 0;
    m_currentParagraph = new Paragraph;
    m_remainingChars = newRemainingChars;
    m_subDocument = newSubDocument;
    m_parsingMode = newParsingMode;

    m_wordDocument->push();
    if ( m_data )
        m_data->push();
}

void Parser97::restoreState()
{
    if ( oldParsingStates.empty() ) {
        std::cerr << "Bug: You messed up the save/restore stack! The stack is empty" << std::endl;
        return;
    }

    if ( m_data )
        m_data->pop();
    m_wordDocument->pop();

    ParsingState ps( oldParsingStates.top() );
    oldParsingStates.pop();

    if ( m_tableRowStart )
        std::cerr << "Bug: We still have to process the table row." << std::endl;
    delete m_tableRowStart;   // Should be a no-op, but I hate mem-leaks even for buggy code ;-)
    m_tableRowStart = ps.tableRowStart;
    m_tableRowLength = ps.tableRowLength;
    m_remainingCells = ps.remainingCells;

    if ( !m_currentParagraph->empty() )
        std::cerr << "Bug: The current paragraph isn't empty." << std::endl;
    delete m_currentParagraph;
    m_currentParagraph = ps.paragraph;

    if ( m_remainingChars != 0 )
        std::cerr << "Bug: Still got " << m_remainingChars << " remaining chars." << std::endl;
    m_remainingChars = ps.remainingChars;
    m_sectionNumber = ps.sectionNumber;

    m_subDocument = ps.subDocument;
    m_parsingMode = ps.parsingMode;
}

U32 Parser97::toLocalCP( U32 globalCP ) const
{
    if ( globalCP < m_fib.ccpText )
        return globalCP;
    globalCP -= m_fib.ccpText;

    if ( globalCP < m_fib.ccpFtn )
        return globalCP;
    globalCP -= m_fib.ccpFtn;

    if ( globalCP < m_fib.ccpHdd )
        return globalCP;
    globalCP -= m_fib.ccpHdd;

    if ( globalCP < m_fib.ccpMcr )
        return globalCP;
    globalCP -= m_fib.ccpMcr;

    if ( globalCP < m_fib.ccpAtn )
        return globalCP;
    globalCP -= m_fib.ccpAtn;

    if ( globalCP < m_fib.ccpEdn )
        return globalCP;
    globalCP -= m_fib.ccpEdn;

    if ( globalCP < m_fib.ccpTxbx )
        return globalCP;
    globalCP -= m_fib.ccpTxbx;

    if ( globalCP < m_fib.ccpHdrTxbx )
        return globalCP;
    globalCP -= m_fib.ccpHdrTxbx;

    std::cerr << "Warning: You aimed " << globalCP << " characters past the end of the text!" << std::endl;
    return globalCP;
}

int Parser97::accumulativeLength( int len, const Parser97::Chunk& chunk )
{
    return len + chunk.m_text.length();
}
